/* /////////////////////////////////////////////////////////////////////////////
 * File:        comstl/errorinfo_desc.hpp (formerly comstl_error.h)
 *
 * Purpose:     errorinfo_desc class for accessing description from the COM error.
 *
 * Created:     19th December 2002
 * Updated:     6th October 2005
 *
 * Home:        http://stlsoft.org/
 *
 * Copyright (c) 2002-2005, Matthew Wilson and Synesis Software
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name(s) of Matthew Wilson and Synesis Software nor the names of
 *   any contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/// \file comstl/errorinfo_desc.hpp
///
/// errorinfo_desc class for accessing description from the COM error.
///
/// \note The contents of this file are not a final form, and <b>will</b> change in
/// future releases.

#ifndef COMSTL_INCL_COMSTL_HPP_ERRORINFO_DESC
#define COMSTL_INCL_COMSTL_HPP_ERRORINFO_DESC

#ifndef STLSOFT_DOCUMENTATION_SKIP_SECTION
# define COMSTL_VER_COMSTL_HPP_ERRORINFO_DESC_MAJOR     0
# define COMSTL_VER_COMSTL_HPP_ERRORINFO_DESC_MINOR     0
# define COMSTL_VER_COMSTL_HPP_ERRORINFO_DESC_REVISION  2
# define COMSTL_VER_COMSTL_HPP_ERRORINFO_DESC_EDIT      3
#endif /* !STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Auto-generation and compatibility
 */

/*
[<[STLSOFT-AUTO:NO-DOCFILELABEL]>]
[<[STLSOFT-AUTO:NO-UNITTEST]>]
*/

/* ////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef COMSTL_INCL_H_COMSTL
# include <comstl.h>
#endif /* !COMSTL_INCL_H_COMSTL */
#ifndef STLSOFT_INCL_STLSOFT_HPP_STRING_ACCESS_FWD
# include <stlsoft/string_access_fwd.hpp>
#endif /* !STLSOFT_INCL_STLSOFT_HPP_STRING_ACCESS_FWD */
#ifdef __STLSOFT_CF_THROW_BAD_ALLOC
# include <stdexcept>
#endif /* __STLSOFT_CF_THROW_BAD_ALLOC */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#ifndef _COMSTL_NO_NAMESPACE
# if defined(_STLSOFT_NO_NAMESPACE) || \
     defined(STLSOFT_DOCUMENTATION_SKIP_SECTION)
/* There is no stlsoft namespace, so must define ::comstl */
namespace comstl
{
# else
/* Define stlsoft::comstl_project */

namespace stlsoft
{

namespace comstl_project
{

# endif /* _STLSOFT_NO_NAMESPACE */
#endif /* !_COMSTL_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

/// \defgroup comstl_error_functions Error Functions
/// \ingroup COMSTL
/// \ingroup functions
/// \brief These functions assist in the handling of COM error info
/// @{

/* /////////////////////////////////////////////////////////////////////////////
 * Classes
 */

class errorinfo_desc
{
/// \name Types
/// @{
public:
    typedef errorinfo_desc  class_type;
/// @}

/// \name Construction
/// @{
public:
    errorinfo_desc()
        : m_description(NULL)
        , m_description_a(NULL)
    {
        IErrorInfo  *pei;

        if(S_OK == ::GetErrorInfo(0, &pei))
        {
            get_description(pei, m_description);

            pei->Release();
        }
    }
    errorinfo_desc(IErrorInfo *pei)
        : m_description(NULL)
        , m_description_a(NULL)
    {
        pei->GetDescription(&m_description);
    }
    errorinfo_desc(errorinfo_desc const &rhs)
        : m_description(::SysAllocString(rhs.m_description))
        , m_description_a(NULL)
    {}

    ~errorinfo_desc()
    {
        ::SysFreeString(m_description);
        ::CoTaskMemFree(m_description_a);
    }
/// @}

/// \name Accessors
/// @{
public:
    LPCOLESTR c_str() const
    {
        return m_description;
    }
    char const *c_str_a() const
    {
        return const_cast<class_type*>(this)->check_description_a_();
    }
/// @}

/// \name Implementation
/// @{
private:
    char const *check_description_a_()
    {
        if( NULL == m_description_a &&
            NULL != m_description)
        {
            int cch = ::WideCharToMultiByte(0, 0, m_description, -1, NULL, 0, NULL, NULL);

            m_description_a = (char*)::CoTaskMemAlloc((1 + cch) * sizeof(char));

            if(NULL == m_description_a)
            {
#ifdef __STLSOFT_CF_THROW_BAD_ALLOC
                throw stlsoft_ns_qual_std(bad_alloc)();
#endif /* __STLSOFT_CF_THROW_BAD_ALLOC */
            }
            else
            {
                ::WideCharToMultiByte(0, 0, m_description, -1, m_description_a, 1 + cch, NULL, NULL);
            }
        }

        return m_description_a;
    }

    void get_description(IErrorInfo *pei, BSTR &bstr)
    {
        if( 0 != pei &&
            SUCCEEDED(pei->GetDescription(&bstr)))
        {
            // Now trim backwards to get rid of any trailing
            // whitespace
            LPOLESTR    begin   =   bstr;
            LPOLESTR    end     =   begin + ::SysStringLen(bstr);
            LPOLESTR    last    =   end;

            for(; begin != last; --last)
            {
                OLECHAR ch = *(last - 1);

                // This is not terribly internationalised, but
                // will suffice for these simple purposes.
                if( ch == ' ' ||
                    ch == '\t' ||
                    ch == '\r' ||
                    ch == '\n')
                {
                }
                else
                {
                    break;
                }
            }

            if(last != end)
            {
                BSTR    description =   m_description;

                if(!::SysReAllocStringLen(&m_description, m_description, static_cast<UINT>(last - begin)))
                {
                    m_description = description;
                }
            }
        }
    }
/// @}

/// \name Members
/// @{
protected:
    BSTR    m_description;
    char    *m_description_a;
/// @}

// Not to be implemented
protected:
    class_type const &operator =(class_type const &);
};

/* ////////////////////////////////////////////////////////////////////////// */

/// @} // end of group comstl_error_functions

/* ////////////////////////////////////////////////////////////////////////// */

#ifndef _COMSTL_NO_NAMESPACE
# if defined(_STLSOFT_NO_NAMESPACE) || \
     defined(STLSOFT_DOCUMENTATION_SKIP_SECTION)
} // namespace comstl
# else
} // namespace comstl_project
} // namespace stlsoft
# endif /* _STLSOFT_NO_NAMESPACE */
#endif /* !_COMSTL_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* !COMSTL_INCL_COMSTL_HPP_ERRORINFO_DESC */

/* ////////////////////////////////////////////////////////////////////////// */
